"""Implementation of Information Utils
"""

import math
import numpy as np
import numba as nb
from typing import Optional

from .hist import (
        hist1D,
        hist2D,
        hist3D)
from .utils import (
        shuffle_bin_array)


@nb.jit(
    cache=True,
    nogil=True,
    nopython=True)
def entropy(
        X: np.ndarray,
        x_bins: int,
        base: int = 2) -> float:
    """Calculates the empirical entropy of an array.
    
    Calculated using the form:

    .. math::
        H(X) = -\sum_{i=1}^{n}{ P( X_{i} ) \log{P(X_{i})} }
        
    
    Parameters
    ----------
    X: np.ndarray
        The unquantized array to calculate entropy over
    x_bins: int
        The number of bins to split the array into
    base: int
        The base of the logarithm

    Returns
    -------
    float
        The calculated entropy: H(X)
    """
    c_x = hist1D(X, x_bins).ravel()
    p_x = c_x / c_x.sum()
    info = 0.
    for i in np.arange(x_bins):
        if p_x[i] == 0:
            continue
        info -= p_x[i] * np.log(p_x[i]) / np.log(base)
    return info


@nb.jit(
    cache=True,
    nogil=True,
    nopython=True)
def joint_entropy(
        X: np.ndarray,
        Y: np.ndarray,
        x_bins: int,
        y_bins: int,
        base: int = 2) -> float:
    """Calculates the joint entropy of two random variables
    
    Calculated using the form:

    .. math::
        H(X,Y) = - \sum_{x∈X}\sum_{y∈Y} P(x,y) \log{P(x,y)}

    Parameters
    ----------
    X: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    Y: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    x_bins: int 
        the number of bins in `X`. equivalent to `max(X) + 1`
    y_bins: int,
        the number of bins in `Y`. equivalent to `max(Y) + 1`
    base: int
        the base of the logarithm

    Returns
    -------
    float
        The calculated joint entropy H(X,Y)
    """
    c_xy = hist2D(X, Y, x_bins, y_bins)
    p_xy = c_xy / c_xy.sum()
    info = 0.
    for x in np.arange(x_bins):
        for y in np.arange(y_bins):
            if p_xy[x][y] == 0:
                continue
            info -= p_xy[x][y] * np.log(p_xy[x][y]) / np.log(base)
    return info


@nb.jit(
    cache=True,
    nogil=True,
    nopython=True)
def joint_entropy_3d(
        X: np.ndarray,
        Y: np.ndarray,
        Z: np.ndarray,
        x_bins: int,
        y_bins: int,
        z_bins: int,
        base: int = 2) -> float:
    """Calculates the joint entropy of two random variables

    Calculated using the form:

    .. math::
        H(X,Y,Z) = - \sum_{x∈X}\sum_{y∈Y}\sum_{z∈Z} P(x,y,z) \log{P(x,y,z)}

    Parameters
    ----------
    X: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    Y: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    Z: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    x_bins: int 
        the number of bins in `X`. equivalent to `max(X) + 1`
    y_bins: int,
        the number of bins in `Y`. equivalent to `max(Y) + 1`
    z_bins: int,
        the number of bins in `Z`. equivalent to `max(Z) + 1`
    base: int
        the base of the logarithm

    Returns
    -------
    float
        The calculated joint entropy
    """
    c_xyz = hist3D(X, Y, Z, x_bins, y_bins, z_bins)
    p_xyz = c_xyz / c_xyz.sum()
    info = 0.
    for x in np.arange(x_bins):
        for y in np.arange(y_bins):
            for z in np.arange(z_bins):
                if p_xyz[x][y][z] == 0:
                    continue
                info -= p_xyz[x][y][z] \
                        * np.log( p_xyz[x][y][z] ) \
                        / np.log(base)

    return info


@nb.jit(
    cache=True,
    nogil=True,
    nopython=True)
def conditional_entropy(
        X: np.ndarray,
        Y: np.ndarray,
        x_bins: int,
        y_bins: int,
        base: int = 2) -> float:
    """Calculates the conditional entropy of two random variables

    Calculated using the form:

    .. math::
        H(X \mid Y) = - \sum_{x∈X,y∈Y} P(x,y) log{ \\frac {P(x,y)} {P(y)} }

    Parameters
    ----------
    X: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    Y: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    x_bins: int 
        the number of bins in `X`. equivalent to `max(X) + 1`
    y_bins: int,
        the number of bins in `Y`. equivalent to `max(Y) + 1`
    base: int
        the base of the logarithm

    Returns
    -------
    float
        The calculated conditional entropy
    """
    c_xy = hist2D(X, Y, x_bins, y_bins)
    c_y = c_xy.sum(axis=0)
    
    p_xy = c_xy / c_xy.sum()
    p_y = c_y / c_y.sum()

    info = 0.
    for x in np.arange(x_bins):
        for y in np.arange(y_bins):
            if p_xy[x][y] == 0:
                continue
            info -= p_xy[x][y] *\
                    np.log( (p_xy[x][y]) / p_y[y] ) \
                    / np.log(base)
    return info


@nb.jit(
    cache=True,
    nogil=True,
    nopython=True)
def mutual_information(
        X: np.ndarray,
        Y: np.ndarray,
        x_bins: int,
        y_bins: int,
        base: int = 2) -> float:
    """Calculates mutual information for two arrays. 
    
    Calculated using the form:

    .. math::
        I(X;Y) = \sum_{x∈X,y∈Y} P_{X,Y}(x,y) \log{ \\frac {P_{X,Y}(x,y)} {P_X(x)P_Y(y)} }

    Parameters
    ----------
    X: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    Y: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    x_bins: int 
        the number of bins in `X`. equivalent to `max(X) + 1`
    y_bins: int,
        the number of bins in `Y`. equivalent to `max(Y) + 1`
    base: int
        the base of the logarithm

    Returns
    -------
    float
        The calculated mutual information
    """
    c_xy = hist2D(X, Y, x_bins, y_bins)
    c_x = c_xy.sum(axis=1)
    c_y = c_xy.sum(axis=0)

    p_xy = c_xy / c_xy.sum()
    p_x = c_x / c_x.sum()
    p_y = c_y / c_y.sum()

    info = 0.
    for x in range(x_bins):
        for y in range(y_bins):
            if p_xy[x][y] == 0:
                continue
            info += p_xy[x][y] * np.log(p_xy[x][y] / (p_x[x] * p_y[y])) / np.log(base)
    
    return info


@nb.jit(
    cache=True,
    nogil=True,
    nopython=True)
def conditional_mutual_information(
        X: np.ndarray,
        Y: np.ndarray,
        Z: np.ndarray,
        x_bins: int,
        y_bins: int,
        z_bins: int,
        base: int = 2) -> float:
    """Calculates conditional mutual information for three arrays.
    
    Calculated using the form:

    .. math::
        I(X;Y \mid Z) = \sum_{x∈X,y∈Y,z∈Z} P_{X,Y,Z}(x,y,z) log { \\frac {P_Z(z)P_{X,Y,Z}(x,y,z)} {P_{X,Z}(x,z)P_{Y,Z}(y,z)}}

    Parameters
    ----------
    X: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    Y: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    Z: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    x_bins: int 
        the number of bins in `X`. equivalent to `max(X) + 1`
    y_bins: int,
        the number of bins in `Y`. equivalent to `max(Y) + 1`
    z_bins: int,
        the number of bins in `Z`. equivalent to `max(Z) + 1`
    base: int
        the base of the logarithm

    Returns
    -------
    float
        The calculated conditional mutual information
    """

    c_xyz = hist3D(X, Y, Z, x_bins, y_bins, z_bins)
    c_xz = hist2D(X, Z, x_bins, z_bins)
    c_yz = hist2D(Y, Z, y_bins, z_bins)
    c_z = hist1D(Z, z_bins).ravel()

    p_xyz = c_xyz / c_xyz.sum()
    p_xz = c_xz / c_xz.sum()
    p_yz = c_yz / c_yz.sum()
    p_z = c_z / c_z.sum()


    info = 0.
    for x in range(x_bins):
        for y in range(y_bins):
            for z in range(z_bins):
                if p_xyz[x][y][z] == 0:
                    continue
                numer = p_z[z] * p_xyz[x][y][z]
                denom = p_xz[x][z] * p_yz[y][z]
                info += p_xyz[x][y][z] * np.log(numer / denom) / np.log(base)
    return info


@nb.jit(
    cache=True,
    nogil=True,
    nopython=True,
    fastmath=True,
    parallel=True)
def calculate_mi_permutations(
        X: np.ndarray, 
        Y: np.ndarray,
        x_bins: int,
        y_bins: int,
        base: int = 2,
        n: int = 10000) -> np.ndarray:
    """calculates the MI for `n` permutations of X

    Parameters
    ----------
    X: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    Y: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    x_bins: int 
        the number of bins in `X`. equivalent to `max(X) + 1`
    y_bins: int,
        the number of bins in `Y`. equivalent to `max(Y) + 1`
    base: int
        the base of the logarithm
    n: int
        the number of permutations to perform (default = 10000)

    Returns
    -------
    np.ndarray 
        The calculated joint entropy for each of the permutations
    """
    permutations = np.zeros(n)
    for idx in nb.prange(n):
        tmp_X = shuffle_bin_array(X)
        permutations[idx] = mutual_information(
                tmp_X, Y, x_bins, y_bins, base=base)
    return permutations

@nb.jit(
    cache=True,
    nogil=True,
    nopython=True,
    fastmath=True,
    parallel=True)
def calculate_cmi_permutations(
        X: np.ndarray,
        Y: np.ndarray,
        Z: np.ndarray,
        x_bins: int,
        y_bins: int,
        z_bins: int,
        base: int = 2,
        n: int = 10000) -> np.ndarray:
    """calculates the MI for `n` permutations of X

    Parameters
    ----------
    X: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    Y: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    Z: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    x_bins: int
        the number of bins in `X`. equivalent to `max(X) + 1`
    y_bins: int,
        the number of bins in `Y`. equivalent to `max(Y) + 1`
    z_bins: int,
        the number of bins in `Z`. equivalent to `max(Z) + 1`
    base: int
        the base of the logarithm
    n: int
        the number of permutations to perform (default = 10000)

    Returns
    -------
    np.ndarray
        The calculated CMI for each of the permutations
    """
    permutations = np.zeros(n)
    for idx in nb.prange(n):
        tmp_X = shuffle_bin_array(X)
        permutations[idx] = conditional_mutual_information(
            tmp_X, Y, Z, x_bins, y_bins, z_bins, base=base)
    return permutations

@nb.jit(
    cache=True,
    nogil=True,
    nopython=True,
    fastmath=True)
def measure_redundancy(
        X: np.ndarray, 
        Y: np.ndarray,
        Z: np.ndarray,
        x_bins: int,
        y_bins: int,
        z_bins: int,
        base: int = 2) -> float:
    """Measures the redundancy of a pathway via a ratio of
    conditional mutual information and mutual information

    calculated using the form:

    .. math::
        R_i = \\frac {I(Y; X \mid Z)} {I(Y;Z)}

    where:
        X: Gene Expression

        Y: Candidate Pathway
        
        Z: Accepted Pathway

    Parameters
    ----------
    X: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    Y: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    Z: np.ndarray
        a 1D array where each value represents the bin index
        for a gene
    x_bins: int 
        the number of bins in `X`. equivalent to `max(X) + 1`
    y_bins: int,
        the number of bins in `Y`. equivalent to `max(Y) + 1`
    z_bins: int,
        the number of bins in `Z`. equivalent to `max(Z) + 1`
    base: int
        the base of the logarithm

    Returns
    -------
    float
        The calculated R-value
    """
    cmi = conditional_mutual_information(
            Y, 
            X, 
            Z, 
            y_bins, 
            x_bins, 
            z_bins,
            base=base)

    mi = mutual_information(
            Y,
            Z,
            y_bins,
            z_bins,
            base=base)

    return cmi / mi




"""Implementation of the PAGE Algorithm
"""

from .io import (
    GeneSets,
    ExpressionProfile)
from .utils import (
    empirical_pvalue,
    hypergeometric_test,
    benjamini_hochberg)
from .information import (
    mutual_information,
    calculate_mi_permutations,
    calculate_cmi_permutations,
    measure_redundancy,
    conditional_mutual_information)
from .heatmap import Heatmap

import numpy as np
import numba as nb
import pandas as pd
from tqdm import tqdm
from typing import Optional


class PAGE:
    """
    Pathway Analysis of Gene Expression [1]_


    Methods
    -------
    run:
        Perform the PAGE algorithm on the provided data

    Attributes
    ----------
    expression: ExpressionProfile
        the provided expression profile
    ontology: GeneSets
        the provided ontology
    shared_genes: np.ndarray
        the shared genes between the `ExpressionProfile` and `GeneOntology`
    exp_bins: np.ndarray
        the expression bins for the intersection of `shared_genes`.
        of shape (`shared_genes`.size, )
    ont_bool: np.ndarray
        the ontology bins for the intersection of `shared_genes`.
        of shape (`num_pathways`, `shared_genes`.size)
    x_bins: int
        the number of expression bins in `exp_bins`
    y_bins: int
        the number of expression bins in `ont_bool`
    num_pathways: int
        the number of pathways in `ontology`
    information: np.ndarray
        the calculated mutual information for each pathway.
        of shape (`num_pathways`, )
    informative: np.ndarray
        a `bool` array representing whether that pathway was considered informative.
        of shape (`num_pathways`, )
    n_shuffle: int
        the number of performed permutation tests
        (`default = 1e4`)
    alpha: float
        the maximum p-value threshold to consider a pathway informative
        with respect to the permuted mutual information distribution
        (`default = 5e-3`)
    k: int
        the number of contiguous uninformative pathways to consider before
        stopping the informative pathway search
        (`default = 20`)
    r: float
        the ratio of the conditional mutual information of a new accepted
        pathway against the mutual information of that pathway against
        all other accepted pathways. Only active when filter_redundant == True.
        (`default = 5.0`)
    base: int
        the base of the logarithm used when calculating entropy
        (`default = 2`)
    filter_redundant: bool
        whether to perform the pathway redundancy search
        (`default = True`)
    n_jobs: int
        The number of parallel jobs to use in the analysis
        (`default = all available cores`)


    Notes
    =====
    .. [1] H. Goodarzi, O. Elemento, S. Tavazoie, "Revealing Global Regulatory Perturbations across Human Cancers." https://doi.org/10.1016/j.molcel.2009.11.016
    """

    def __init__(
            self,
            expression: ExpressionProfile,
            genesets: GeneSets,
            n_shuffle: int = 1e3,
            alpha: float = 1e-2,
            k: int = 10,
            filter_redundant: bool = False,
            n_jobs: Optional[int] = 1,
            function: Optional[str] = 'cmi',
            redundancy_ratio: Optional[float] = .1):
        """
        Initialize object

        Parameters
        ----------
        expression: ExpressionProfile
            The provided `ExpressionProfile`

        genesets: GeneSets
            the provided `GeneOntology`

        n_shuffle: int
            the number of performed permutation tests
            (`default = 1e4`)

        alpha: float
            the maximum p-value threshold to consider a pathway informative
            with respect to the permuted mutual information distribution
            (`default = 5e-2`)

        k: int
            the number of contiguous uninformative pathways to consider before
            stopping the informative pathway search
            (`default = 20`)

        base: int
            the base of the logarithm used when calculating entropy
            (`default = 2`)

        filter_redundant: bool
            whether to perform the pathway redundancy search
            (`default = True`)

        n_jobs: int
            The number of parallel jobs to use in the analysis
            (`default = all available cores`)
        """

        self.expression = expression
        self.ontology = genesets

        self.n_shuffle = int(n_shuffle)
        self.alpha = float(alpha)
        self.k = int(k)
        self.base = 2
        self.filter_redundant = filter_redundant
        self.n_jobs = n_jobs
        self._set_jobs()
        self.function = function
        self.redundancy_ratio = redundancy_ratio

        if not self.expression.modified and not self.ontology.modified:
            self._intersect_genes()
            self._subset_matrices()
            self._set_sizes()

    def _set_jobs(self):
        """Sets the number of available jobs for numba parallel
        """
        if self.n_jobs:
            self.n_jobs = int(self.n_jobs)
            nb.set_num_threads(self.n_jobs)
        else:
            # default to using all available threads
            nb.set_num_threads(nb.config.NUMBA_NUM_THREADS)

    def _intersect_genes(self):
        """Intersects to genes found in both sets
        """
        self.shared_genes = np.sort(np.intersect1d(
                                    self.expression.genes,
                                    self.ontology.genes))

    def _subset_matrices(self):
        """Subsets the bool arrays to the gene intersection
        """
        self.exp_bins = self.expression.get_gene_subset(self.shared_genes)
        self.ont_bool = self.ontology.get_gene_subset(self.shared_genes)
        self.membership_bins = self.ontology.get_membership_subset(self.shared_genes)

    def _set_sizes(self):
        """Sets the number of bins for the expression and ontology
        as well as the number of pathways found
        """
        self.x_bins = self.exp_bins.max() + 1
        self.y_bins = self.ont_bool.max() + 1
        self.z_bins = self.membership_bins.max() + 1
        self.num_pathways = self.ont_bool.shape[0]

    def _calculate_information(self) -> np.ndarray:
        """Calculates mutual or conditional mutual information for each pathway
        """
        information = np.zeros(self.num_pathways)
        if self.function == 'mi':
            desc = "calculating mutual information"
        else:
            desc = "calculating conditional mutual information"
        pbar = tqdm(range(self.num_pathways), desc=desc)
        for idx in pbar:
            if self.function == 'mi':
                information[idx] = mutual_information(
                    self.exp_bins,
                    self.ont_bool[idx],
                    self.x_bins,
                    self.y_bins,
                    base=self.base)
            else:
                information[idx] = conditional_mutual_information(
                    self.exp_bins,
                    self.ont_bool[idx],
                    self.membership_bins,
                    self.x_bins,
                    self.y_bins,
                    self.z_bins,
                    base=self.base)
        return information

    def _calculate_information_2D(self) -> np.ndarray:
        """Calculates mutual or conditional mutual information for each pathway
        """
        information = np.zeros((self.exp_bins.shape[0], self.num_pathways))
        if self.function == 'mi':
            desc = "calculating mutual information"
        else:
            desc = "calculating conditional mutual information"
        pbar = tqdm(range(self.exp_bins.shape[0]), desc=desc)
        for exp_idx in pbar:
            for idx in range(self.num_pathways):
                if self.function == 'mi':
                    information[exp_idx, idx] = mutual_information(
                        self.exp_bins[exp_idx],
                        self.ont_bool[idx],
                        self.x_bins,
                        self.y_bins,
                        base=self.base)
                else:
                    information[exp_idx, idx] = conditional_mutual_information(
                        self.exp_bins[exp_idx],
                        self.ont_bool[idx],
                        self.membership_bins,
                        self.x_bins,
                        self.y_bins,
                        self.z_bins,
                        base=self.base)
        information = pd.DataFrame(information, columns=self.ontology.pathways)

        return information

    def _calculate_enrichment(self) -> (np.ndarray, np.ndarray):
        """
        Iterates through informative pathways to calculate hypergeometric pvalues
        """
        overrep_pvals = np.zeros((self.pathway_indices.size, self.x_bins))
        underrep_pvals = np.zeros_like(overrep_pvals)

        pbar = tqdm(enumerate(self.pathway_indices), desc="hypergeometric tests")
        for idx, info_idx in pbar:
            pvals = hypergeometric_test(
                self.exp_bins,
                self.ont_bool[info_idx])
            overrep_pvals[idx, :] = pvals[0]
            underrep_pvals[idx, :] = pvals[1]

        return overrep_pvals, underrep_pvals

    def _calculate_informative(self) -> (np.ndarray, np.ndarray):
        """Calculates the informative categories
        """
        n_break = 0
        informative = np.zeros_like(self.information)
        pvalues = np.ones_like(self.information)

        # iterate through most informative pathways
        pbar = tqdm(np.argsort(self.information)[::-1], desc="permutation testing")
        for idx in pbar:

            # calculate mutual information of random permutations
            if self.function == 'mi':
                permutations = calculate_mi_permutations(
                    self.exp_bins,
                    self.ont_bool[idx],
                    self.x_bins,
                    self.y_bins,
                    n=self.n_shuffle)
            else:
                permutations = calculate_cmi_permutations(
                    self.exp_bins,
                    self.ont_bool[idx],
                    self.membership_bins,
                    self.x_bins,
                    self.y_bins,
                    self.z_bins,
                    n=self.n_shuffle)

            # calculate empirical pvalue against randomization

            pvalues[idx] = empirical_pvalue(
                permutations,
                self.information[idx])

            if pvalues[idx] > self.alpha:
                n_break += 1
                if n_break == self.k:
                    break
            else:
                informative[idx] = 1
                n_break = 0

        return (informative, pvalues)

    def _consolidate_pathways(self) -> np.ndarray:
        """Consolidate redundant pathways
        """
        existing = []
        inf_idx = np.flatnonzero(self.informative)

        # iterate through cmi in descending order
        pbar = tqdm(np.argsort(self.information)[::-1], desc="consolidating redundant pathways")
        for idx in pbar:

            # skip indices that are not informative
            if idx not in inf_idx:
                continue

            # if there are no existing pathways yet start the chain
            if len(existing) == 0:
                existing.append(idx)
                continue

            # initialize reduncancy information array
            all_ri = np.zeros(len(existing))

            # calculate redundancy
            for i, e in enumerate(existing):
                all_ri[i] = measure_redundancy(
                    self.exp_bins,
                    self.ont_bool[idx],
                    self.ont_bool[e],
                    self.x_bins,
                    self.y_bins,
                    self.y_bins)

            if all(all_ri > self.redundancy_ratio):
                existing.append(idx)
            else:
                pass

        return np.array(existing)

    def _gather_results(self) -> pd.DataFrame:
        """Gathers the results from the experiment into a single dataframe
        """
        # estimate sign
        self.log_overrep_pvals = np.log10(self.overrep_pvals)
        self.log_underrep_pvals = np.log10(self.underrep_pvals)
        self.graphical_ar = np.minimum(self.log_overrep_pvals, self.log_underrep_pvals)
        self.graphical_ar[self.log_overrep_pvals < self.log_underrep_pvals] *= -1  # make overrepresented positive
        n_bins = self.graphical_ar.shape[1]
        s1 = self.graphical_ar[:, :n_bins // 3].copy()
        s2 = self.graphical_ar[:, -n_bins // 3:].copy()
        s1[s1 < 0] = 0
        s2[s2 < 0] = 0
        sign = s1.sum(1) <= s2.sum(1)
        sign = sign.astype(int)
        sign[sign == 0] = -1
        results = pd.DataFrame({"pathway": self.ontology.pathways[self.pathway_indices],
                                "CMI": self.information[self.pathway_indices],
                                "p-value": self.pvalues[self.pathway_indices],
                                "Regulation pattern": sign}
                               )
        return results

    def _make_heatmap(self):
        """

        Returns
        -------

        """

        hm = Heatmap(np.array(self.results['pathway']),
                     self.graphical_ar)

        hm.add_gene_expression(self.expression.genes, self.expression.raw_expression)
        return hm

    def run(self) -> (pd.DataFrame, Heatmap):
        """
        Perform the PAGE algorithm

        Returns
        =======
        pd.DataFrame
            A dataframe representing the results of the analysis

        Examples
        ========
        >>> p = PAGE(exp, ont)
        >>> p.run()
        """

        # calculate mutual information
        if len(self.exp_bins.shape) == 2:
            self.information = self._calculate_information_2D()
            return self.information

        self.information = self._calculate_information()

        # select informative pathways
        self.informative, self.pvalues = self._calculate_informative()

        # filter redundant pathways
        if self.filter_redundant:
            self.pathway_indices = self._consolidate_pathways()
        else:
            self.pathway_indices = np.flatnonzero(self.informative)
        if len(self.pathway_indices) != 0:
            # hypergeometric testing over selected pathways
            self.overrep_pvals, self.underrep_pvals = self._calculate_enrichment()

            self.results = self._gather_results()
            self.hm = self._make_heatmap()

            return self.results, self.hm
        else:
            return pd.DataFrame(columns=["pathway", "CMI", "p-value", "Regulation pattern"]), None

    def get_enriched_genes(self, name) -> list:
        assert name in self.ontology.pathways, "pathway not present"
        pathway_idx = np.where(self.ontology.pathways == name)[0][0]
        pathway_binary = self.ont_bool[pathway_idx]
        res = []
        for bin in set(self.exp_bins):
            genes_in_bin = self.shared_genes[np.where(pathway_binary & (self.exp_bins == bin))[0]]
            res.append(genes_in_bin)
        return res



"""Testing Expression IO 
"""


import pandas as pd
import numpy as np
from pypage import ExpressionProfile


N_GENES=1000
N_BINS=5
T = 100


def get_expression() -> (np.ndarray, np.ndarray):
    genes = np.array([f"g.{g}" for g in np.arange(N_GENES)])
    scores = np.random.normal(size=N_GENES)
    return genes, scores


def get_bins() -> (np.ndarray, np.ndarray):
    genes = np.array([f"g.{g}" for g in np.arange(N_GENES)])
    bins = np.random.choice(N_BINS, size=N_GENES)
    return genes, bins


def test_load_expression_nobins():
    for _ in np.arange(T):
        genes, scores = get_expression()
        exp = ExpressionProfile(genes, scores)
        assert exp.n_genes == N_GENES
        assert exp.n_bins == 10


def test_load_expression():
    for _ in np.arange(T):
        genes, expression = get_expression()
        exp = ExpressionProfile(genes, expression, n_bins=N_BINS)
        assert exp.n_genes == N_GENES
        assert exp.n_bins == N_BINS


def test_load_expression_assertion():
    for i in np.arange(T):
        genes, expression = get_expression()

        if i % 2 == 0:
            genes = genes[np.random.random(N_GENES) < 0.3]
        else:
            expression = expression[np.random.random(N_GENES) < 0.3]

        try:
            ont = ExpressionProfile(genes, expression)
        except AssertionError:
            continue

        assert False


"""def test_load_expression_strategy_hist():
    for _ in np.arange(T):
        genes, expression = get_expression()

        # test hist binning
        hist_exp = ExpressionProfile(
                genes, 
                expression, 
                n_bins=N_BINS, 
                bin_strategy='hist')
        
        assert hist_exp.n_genes == N_GENES
        assert hist_exp.n_bins == N_BINS

        # test split binning
        split_exp = ExpressionProfile(
                genes, 
                expression, 
                n_bins=N_BINS, 
                bin_strategy='split')
        
        assert split_exp.n_genes == N_GENES
        assert split_exp.n_bins == N_BINS

        # assert bins are different
        assert np.any(hist_exp.bin_sizes != split_exp.bin_sizes)"""


def test_load_bins():
    genes, bins = get_bins()
    exp = ExpressionProfile(genes, bins, n_bins=N_BINS)
    assert exp.n_genes == N_GENES
    assert exp.n_bins == N_BINS


def test_expression_conversion():
    df = pd.read_csv('example_data/bladder_refseq.tsv.gz',
                     sep="\t",
                     header=0,
                     names=["gene", "exp"],
                     compression='gzip')
    exp = ExpressionProfile(df.iloc[:, 0],
                            df.iloc[:, 1],
                            n_bins=20)
    exp.convert_from_to('refseq', 'ensg', 'human')



"""Data input and output handling for ontologies
"""


import sys
import numpy as np
from typing import Optional
from .accession_types import change_accessions
import gzip


class GeneSets:
    """Container for a Set of Pathways

    Attributes
    ==========
    genes: np.ndarray
        the sorted list of genes found in the pathways
    pathways: np.ndarray
        the sorted list of pathways found in the index
    n_genes: int
        the number of genes found
    n_pathways: int
        the number of pathways found
    bool_array: np.ndarray
        a (n_pathways, n_genes) bit array representing the bin-position of each gene
    pathway_sizes: np.ndarray
        a (n_pathways, ) array where each value represents the number of genes in that pathway index
    avg_p_size: float
        the mean number of genes across pathways
    membership: np.ndarray
        array representing the membership of each gene in an annotation
    n_bins: int
        number of bins to use when binning membership array
    bin_array: np.ndarray
        binned membership array

    Methods
    -------
    get_gene_subset:
        returns a subset of the `bool_array` for the provided gene list
    filter_pathways:
        filters pathways based on membership
    """
    def __init__(
            self,
            genes: Optional[np.ndarray] = None,
            pathways: Optional[np.ndarray] = None,
            ann_file: Optional[str] = None,
            n_bins: Optional[int] = 3,
            first_col_is_genes: Optional[bool] = False):
        """
        Parameters
        ==========
        genes: np.ndarray
            an array of gene names
        pathways: np.ndarray
            an array associated pathways
        """
        self.modified = False
        if ann_file:
            self._read_annotation_file(ann_file, first_col_is_genes)
        else:
            self._validate_inputs(genes, pathways)
            self._load_genes(genes)
            self._load_pathways(pathways)
            self._build_bool_array(genes, pathways)
        self._make_membership_profile()
        self.n_bins = n_bins

    def _read_annotation_file(self,
                              ann_file: str,
                              first_col_is_genes: Optional[bool] = False):
        """
         Reads annotation files which are in a descriptive format,
         i.e, "Gene1 Pathway1 Pathway2..." or "Pathway1 Gene1 Gene2..."

         Parameters
         ----------
         ann_file
             a file name of the annotation
         first_is_gene
             a parameter which specifies whether a gene is
         Returns
         -------
         pd.DataFrame
             a dataframe in a long format
         """
        if ann_file[-2:] == 'gz':
            f = gzip.open(ann_file, 'r')
        else:
            f = open(ann_file)

        row_names = []
        column_names = set()
        for line in f:
            if ann_file[-2:] == 'gz':
                line = line.decode('UTF-8')
            els = line.rstrip().split('\t')
            row_names.append(els[0])
            els.pop(0)
            if 'http://' in els[0]:
                els.pop(0)
            column_names |= set(els)
        column_names = list(column_names)

        positions = dict(zip(column_names, np.arange(len(column_names))))
        db_profiles = np.zeros((len(row_names), len(column_names)), dtype=int)

        if ann_file[-2:] == 'gz':
            f = gzip.open(ann_file, 'r')
        else:
            f = open(ann_file)

        i = 0
        for line in f:
            if ann_file[-2:] == 'gz':
                line = line.decode('ASCII')
            els = line.rstrip().split('\t')[1:]
            if 'http://' in els[0]:
                els.pop(0)
            indices = [positions[el] for el in els]
            db_profiles[i, indices] = 1
            i += 1

        if first_col_is_genes:
            db_profiles = db_profiles.T
            db_genes = row_names
            db_names = column_names
        else:
            db_genes = column_names
            db_names = row_names
        self.pathways = np.array(db_names)
        self.genes = np.array(db_genes)
        self.bool_array = db_profiles
        self.n_genes = len(self.genes)
        self.n_pathways = len(self.pathways)

    def _validate_inputs(
            self,
            x: np.ndarray,
            y: np.ndarray):
        """validates inputs are as expected
        """
        assert x.size > 0, \
            "provided array must not be empty"
        assert x.size == y.size, \
            "genes and pathway arrays must be equal sized"
        assert x.shape == y.shape, \
            "genes and pathway arrays must be equally shaped"

    def _load_genes(
            self,
            genes: np.ndarray):
        """load genes and associated indices
        """
        self._gene_indices = {n: idx for idx, n in enumerate(np.sort(np.unique(genes)))}
        self.genes = np.array(list(self._gene_indices.keys()))
        self.n_genes = self.genes.size

    def _load_pathways(
            self,
            pathways: np.ndarray):
        """load pathways and associated indices
        """
        self._pathway_indices = {n: idx for idx, n in enumerate(np.sort(np.unique(pathways)))}
        self.pathways = np.array(list(self._pathway_indices.keys()))
        self.n_pathways = self.pathways.size

    def _calculate_pathway_sizes(self):
        """calculates and sets pathway sizes
        """
        self.pathway_sizes = self.bool_array.sum(axis=1)
        self.avg_p_size = self.pathway_sizes.mean()

    def _build_bool_array(
            self,
            genes: np.ndarray,
            pathways: np.ndarray):
        """create the bool array of genes/pathway interactions
        """
        self.bool_array = np.zeros((self.n_pathways, self.n_genes), dtype=int)
        for g, p in zip(genes, pathways):
            self.bool_array[self._pathway_indices[p]][self._gene_indices[g]] += 1

        self._calculate_pathway_sizes()

        # these may change with filtering so better to remove them
        del self._pathway_indices
        del self._gene_indices

    def _build_bin_split(
            self,
            membership: np.ndarray,
            n_bins: int) -> np.ndarray:
        """converts membership array to binned data using equivlanet split method
        """
        argidx = np.argsort(membership)
        max_size = membership.size
        bin_size = int(max_size / n_bins)

        bin_identities = np.zeros(max_size, dtype=int)
        self.bin_sizes = np.zeros(n_bins, dtype=int)
        self.bin_ranges = np.zeros(n_bins)
        self.bin_ranges[-1] = membership.max()

        for i in np.arange(0, n_bins):
            lower_bound = membership[argidx[bin_size * i]]

            if i < n_bins - 1:
                upper_bound = membership[argidx[bin_size * (i + 1)]]
                mask = (membership >= lower_bound) & (membership < upper_bound)

            # put remaining into last bin
            else:
                mask = (membership >= lower_bound)

            self.bin_sizes[i] = mask.sum()
            self.bin_ranges[i] = lower_bound
            bin_identities[mask] = i

        return bin_identities

    def _make_membership_profile(self) -> np.ndarray:
        """create a gene membership array"""
        self.membership = self.bool_array.sum(0)

    def get_gene_subset(
            self,
            gene_subset: np.ndarray) -> np.ndarray:
        """
        Index the bool-array for the required gene subset.
        Expects the subset to be sorted

        Parameters
        ==========
        gene_subset: np.ndarray
            a list of genes to subset the bin_array to

        Returns
        =======
        np.ndarray
            the bool_array subsetted to the indices of the `gene_subset`
        """
        idxs = [np.where(self.genes == gene)[0][0] for gene in gene_subset]
        self.sub_bool_array = self.bool_array[:, idxs]
        self.modified = True
        return self.sub_bool_array

    def get_membership_subset(
            self,
            gene_subset: np.ndarray) -> np.ndarray:
        """
        Index the bool-array for the required gene membership subset and bin it.
        Expects the subset to be sorted

        Parameters
        ==========
        gene_subset: np.ndarray
            a list of genes to subset the bin_array to

        Returns
        =======
        np.ndarray
            the bool_array subsetted to the indices of the `gene_subset`
        """
        idxs = [np.where(self.genes == gene)[0][0] for gene in gene_subset]
        sub_membership = self.membership[idxs]
        sub_membership_binned = self._build_bin_split(sub_membership, self.n_bins)
        return sub_membership_binned

    def filter_pathways(
            self,
            min_size: Optional[int] = None,
            max_size: Optional[int] = None):
        """
        Filters pathways to those within a specified range

        Parameters
        ----------
        min_size: Optional[int]
            the minimum size of the pathways, defaults to 0
        max_size: Optional[int]
            the maximum size of the pathways, default to current maximum
        """
        if not min_size and not max_size:
            print("No minimum or maximum size provided. Doing nothing.", file=sys.stderr)
            return
        if not min_size:
            min_size = 0
        if not max_size:
            max_size = np.max(self.pathway_sizes)
        assert min_size >= 0, "Provided minimum must be >= 0"
        assert max_size > min_size, f"Provided maximum must be > min_size: {min_size}"

        # determine pathway-level mask
        p_mask = (self.pathway_sizes >= min_size) & (self.pathway_sizes <= max_size)

        # filter pathways
        self.bool_array = self.bool_array[p_mask]
        self.pathway_sizes = self.pathway_sizes[p_mask]
        self.pathways = self.pathways[p_mask]

        # determine gene level mask
        g_mask = self.bool_array.sum(axis=0) != 0

        # filter genes with no pathways
        self.bool_array = self.bool_array[:, g_mask]
        self.genes = self.genes[g_mask]

    def convert_from_to(self,
                        input_format: str,
                        output_format: str,
                        species: Optional[str] = 'human'):
        """
        A function which changes accessions
        Parameters
        ----------
        input_format
            input accession type, takes 'enst', 'ensg', 'refseq', 'entrez', 'gs', 'ext'
        output_format
            output accession type, takes 'enst', 'ensg', 'refseq', 'entrez', 'gs', 'ext'
        species
            analyzed species, takes either 'human' or 'mouse'
        """
        if input_format in ['ensg', 'enst', 'refseq']:  # for ex., remove '.1' from 'ENSG00000128016.1'
            self.genes = np.array([gene.split('.')[0] for gene in self.genes])
        self.genes = change_accessions(self.genes,
                                       input_format,
                                       output_format,
                                       species)

    def reset(self):
        self.modified = False

    def __repr__(self) -> str:
        """
        """
        s = ""
        s += "Gene Ontology\n"
        s += f">> num_genes: {self.n_genes}\n"
        s += f">> num_pathways: {self.n_pathways}\n"
        return s





"""Testing an example usage of the PAGE algorithm after its upgrade
"""

import pytest
import pandas as pd
from pypage import (
    PAGE,
    ExpressionProfile,
    GeneSets)


@pytest.fixture()
def load_expression():
    df = pd.read_csv('example_data/bladder_refseq.tsv.gz',
                     sep="\t",
                     header=0,
                     names=["gene", "exp"])
    exp = ExpressionProfile(df.iloc[:, 0],
                             df.iloc[:, 1],
                             n_bins=10)
    exp.convert_from_to('refseq', 'ensg', 'human')
    return exp


@pytest.fixture()
def load_ontology():
    ont = GeneSets(ann_file='example_data/hg38_cistrome_index.txt.gz', n_bins=6)
    return ont


def test_run(load_expression, load_ontology):
    p = PAGE(
        load_expression,
        load_ontology,
        n_shuffle=100,
        k=7,
        filter_redundant=True
        )
    results, hm = p.run()
    # print(results)
    hm.convert_from_to('gs', 'ensg', 'human')
    hm.save('test_heatmap', show_reg=True)



import pybiomart
import numpy as np


def change_accessions(ids: np.ndarray,
                      input_format: str,
                      output_format: str,
                      species: str) -> np.ndarray:
    """
    A function which changes accessions
    Parameters
    ----------
    ids
    input_format
        input accession type, takes 'enst', 'ensg', 'refseq', 'entrez', 'gs', 'ext'
    output_format
        output accession type, takes 'enst', 'ensg', 'refseq', 'entrez', 'gs', 'ext'
    species
        analyzed species, takes either 'human' or 'mouse'
    Returns
    -------

    """
    if input_format == output_format:
        return ids
    else:
        if species == 'mouse':
            dataset = pybiomart.Dataset(name='mmusculus_gene_ensembl', host='http://www.ensembl.org')
        elif species == 'human':
            dataset = pybiomart.Dataset(name='hsapiens_gene_ensembl', host='http://www.ensembl.org')
        # print(*dataset.attributes.keys(), sep='\n')
        mart_attributes = {'enst': ['ensembl_transcript_id'], 'ensg': ['ensembl_gene_id'],
                           'refseq': ['refseq_mrna', 'refseq_mrna_predicted', 'refseq_ncrna',
                                      'refseq_ncrna_predicted'], 'entrez': ['entrezgene_id'],
                           'gs': ['entrezgene_accession'], 'ext': ['external_gene_name']}
        input_to_output = {}
        output_attributes = mart_attributes[output_format]
        if output_format == 'refseq':
            output_attributes = [output_attributes[0]]
        for mart in mart_attributes[input_format]:
            df1 = dataset.query(attributes=[mart] + output_attributes)
            df1 = df1[df1.iloc[:, 0].notna()]
            df1 = df1[df1.iloc[:, 1].notna()]
            if input_format == 'entrez' or output_format == 'entrez':
                df1['NCBI gene ID'] = df1['NCBI gene ID'].apply(lambda x: '%.f' % x)
            if input_format == 'gene_symbol' or output_format == 'gene_symbol':
                upper = lambda x: x.upper() if type(x) == str else x
                df1['NCBI gene accession'] = df1['NCBI gene accession'].apply(upper)
            input_to_output = {**input_to_output, **dict(zip(df1.iloc[:, 0], df1.iloc[:, 1]))}

        new_ids = []
        for id_ in ids:
            if id_ in input_to_output.keys():
                new_ids.append(input_to_output[id_])
            else:
                new_ids.append('-')
        return np.array(new_ids)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import matplotlib
from typing import Optional
import copy
from .io.accession_types import change_accessions


class Heatmap:
    """
    An object which is used to build a heatmap, representing pathway deregulation
    Attributes
    ==========
    pathways: np.ndarray
    graphical_ar: np.ndarray
    regulator_exp: np.ndarray
    cmap_main: str
    cmap_reg: str
    n_pathways: int
    n_bins: int
    isreg: bool
    fig: matplotlib.figure.Figure
    ax: matplotlib.figure.Axes
    ==========
    Methods
    ----------
    save
        saves a heatmap
    show
        shows a heatmap
    """
    def __init__(self,
                 pathways: np.ndarray,
                 graphical_ar: np.ndarray,
                 regulator_exp: Optional[np.ndarray] = None,
                 cmap_main: Optional[str] = 'viridis',
                 cmap_reg: Optional[str] = 'plasma'):
        self.pathways = pathways
        self.graphical_ar = graphical_ar
        self.regulator_exp = regulator_exp
        self.cmap_main = cmap_main
        self.cmap_reg = cmap_reg
        self.n_pathways = self.graphical_ar.shape[0]
        self.n_bins = self.graphical_ar.shape[1]
        self.expression = None
        self.genes = None
        self.isreg = False

        self.regulator_names = [gene.split('_')[0] for gene in self.pathways]

    def _subset_and_sort_pathways(self,
                                  max_rows: int = 50) -> (np.ndarray, np.ndarray):
        """

        Parameters
        ----------
        max_rows
            maximum number of rows to use on a heatmap
        Returns
        -------

        """
        upregulated_mask = self.graphical_ar[:, :self.n_bins // 2].sum(1) <= self.graphical_ar[:, self.n_bins // 2:].sum(1)
        upregulated_idxs = np.where(upregulated_mask)[0]
        downregulated_idxs = np.where(~upregulated_mask)[0]
        if 0 < max_rows < self.n_pathways:
            additional_up = max(0, max_rows//2 - len(downregulated_idxs))
            additional_down = max(0, max_rows//2 - len(upregulated_idxs))
            upregulated_idxs = upregulated_idxs[:max_rows//2 + additional_up]
            downregulated_idxs = downregulated_idxs[:max_rows//2 + additional_down]
        sub_idxs = np.concatenate((upregulated_idxs, downregulated_idxs))

        sub_graphical_ar = self.graphical_ar[sub_idxs]
        sub_pathways = self.pathways[sub_idxs]
        if self.isreg:
            sub_regulator_exp = self.regulator_exp[sub_idxs]

        sort_idxs = np.arange(sub_graphical_ar.shape[0])
        for i in range(self.n_bins-2):
            sort_idxs = sorted(sort_idxs, key=lambda x: sum(sub_graphical_ar[x][i:i+3]), reverse=True)
        sorted_pathways = sub_pathways[sort_idxs]
        sorted_graphical_ar = sub_graphical_ar[sort_idxs]
        if self.isreg:
            sorted_regulator_exp = sub_regulator_exp[sort_idxs]
        else:
            sorted_regulator_exp = None

        return sorted_pathways, sorted_graphical_ar, sorted_regulator_exp

    def _columnwise_heatmap(self,
                            graphical_ar: np.ndarray,
                            regulator_exp: np.ndarray,
                            max_val: int):
        self.ims = []
        if self.isreg:
            current_cmap = copy.copy(matplotlib.cm.get_cmap(self.cmap_reg))
            current_cmap.set_bad(color='black')
            im = self.ax[0].imshow(np.atleast_2d(regulator_exp).T, cmap=current_cmap, aspect="auto", vmin=-1, vmax=1)
            self.ims.append(im)
            im = self.ax[1].imshow(np.atleast_2d(graphical_ar),
                                   cmap=self.cmap_main, aspect="auto", vmin=-max_val, vmax=max_val)
            self.ims.append(im)
        else:
            im = self.ax.imshow(np.atleast_2d(graphical_ar),
                                cmap=self.cmap_main, aspect="auto", vmin=-max_val, vmax=max_val)
            self.ims.append(im)

    def _add_colorbar(self):
        self.fig.subplots_adjust(left=0.06, right=0.65)
        rows = 1 + self.isreg
        cols = 1
        gs = GridSpec(rows, cols)
        gs.update(left=0.8, right=0.85, wspace=1, hspace=0.3, bottom=0.2, top=0.5)

        if self.isreg:
            colorbar_names = ['Regulator\'s \n expression', 'Regulon\'s \n enrichment']
            colorbar_images = [0, 1]
        else:
            colorbar_names = ['Regulon\'s \n enrichment']
            colorbar_images = [-1]

        for i in colorbar_images:
            cax = self.fig.add_subplot(gs[i // cols, i % cols])
            self.fig.colorbar(self.ims[i], cax=cax)
            cax.set_title(colorbar_names[i], fontsize=10)

    def _make_heatmap(self, max_rows, max_val, title):
        pathways, graphical_ar, regulator_exp = self._subset_and_sort_pathways(max_rows)
        n_pathways = len(pathways)
        plt.rcParams.update({'font.weight': 'roman'})
        plt.rcParams.update({'ytick.labelsize': 10})
        fontsize_pt = plt.rcParams['ytick.labelsize']
        dpi = 72.27
        matrix_height_pt = (fontsize_pt + 30 / 2) * n_pathways
        matrix_height_in = matrix_height_pt / dpi
        matrix_width_pt = (fontsize_pt + 50 / 2) * self.n_bins
        matrix_width_in = matrix_width_pt / dpi
        figure_height = matrix_height_in
        figure_width = matrix_width_in

        if self.isreg:
            self.fig, self.ax = plt.subplots(1, 2, figsize=(figure_width, figure_height),
                                             gridspec_kw={'width_ratios': [1, self.n_bins]})
            self.fig.subplots_adjust(wspace=0.05)
        else:
            self.fig, self.ax = plt.subplots(1, 1, figsize=(figure_width, figure_height))

        self._columnwise_heatmap(graphical_ar, regulator_exp, max_val)

        if self.isreg:
            self.ax[0].set(xticks=[], yticks=np.arange(n_pathways), yticklabels=pathways)
            self.ax[0].xaxis.set_label_position('top')
            self.ax[1].set(xticks=[], yticks=[])
            self.ax[1].xaxis.set_label_position('top')
        else:
            self.ax.set(xticks=[], yticks=np.arange(n_pathways), yticklabels=pathways)
            self.ax.xaxis.set_label_position('top')
            plt.xticks(rotation=90)
        plt.title(title)
        self._add_colorbar()

    def add_gene_expression(self,
                            genes: np.ndarray,
                            expression: np.ndarray):
        self.expression = expression
        self.genes = genes

    def _calculate_regulator_expression(self):
        self.regulator_exp = np.array([np.nan] * len(self.regulator_names))
        for i, gene in enumerate(self.regulator_names):
            if gene in self.genes:
                self.regulator_exp[i] = self.expression[np.where(self.genes == gene)[0][0]]

    def save(self,
             output_name: str,
             max_rows: Optional[int] = 50,
             show_reg: Optional[bool] = False,
             max_val: Optional[int] = 5,
             title=''):
        """
        saves a heatmap
        Parameters
        ----------
        max_rows
            maximum number of rows in a heatmap, if -1 provided then the number of rows equals the number of pathways
        output_name: str
            name of the output file, if extension provided the file is saved using that extension,
            if not the picture is saved using svg format
        Returns
        -------
        creates a file containing a picture of the heatmap
        """
        if self.expression is not None:
            self._calculate_regulator_expression()
            self.isreg = show_reg and (sum(~np.isnan(self.regulator_exp)) != 0)
        self._make_heatmap(max_rows, max_val, title)
        if output_name.split('.')[-1] not in ['svg', 'jpg', 'png']:
            output_name += '.svg'
        self.fig.savefig(output_name, bbox_inches='tight')

    def show(self,
             max_rows: Optional[int] = 50,
             show_reg: Optional[bool] = False,
             max_val: Optional[int] = 5,
             title=''):
        """
        shows a heatmap
        Returns
        -------
        """
        if self.expression is not None:
            self._calculate_regulator_expression()
            self.isreg = show_reg and (sum(~np.isnan(self.regulator_exp)) != 0)
        self._make_heatmap(max_rows, max_val, title)
        self.fig.show()

    def convert_from_to(self,
                        input_format: str,
                        output_format: str,
                        species: Optional[str] = 'human'):
        """
        A function which changes accessions
        Parameters
        ----------
        input_format
            input accession type, takes 'enst', 'ensg', 'refseq', 'entrez', 'gs', 'ext'
        output_format
            output accession type, takes 'enst', 'ensg', 'refseq', 'entrez', 'gs', 'ext'
        species
            analyzed species, takes either 'human' or 'mouse'
        """
        if input_format in ['ensg', 'enst', 'refseq']:  # for ex., remove '.1' from 'ENSG00000128016.1'
            self.regulator_names = np.array([gene.split('.')[0] for gene in self.regulator_names])
        self.regulator_names = change_accessions(self.regulator_names,
                                                 input_format,
                                                 output_format,
                                                 species)



"""Histogram functions
"""

import numpy as np
import numba as nb


@nb.jit(
    nopython=True, 
    nogil=True, 
    cache=True, 
    fastmath=True)
def hist1D(
        arr_a: np.ndarray, 
        bin_a: int) -> np.ndarray:
    """calculates the number of events in all bins 

    Parameters
	----------
    arr_a: np.ndarray
        a 1D array where each value represents a the cluster
        identity of a specific gene
    bin_a: np.ndarray
        the number of clusters in array 1. Should be equivalent
        to the maximum value in `arr_a` + 1

    Returns
	-------
    np.ndarray
        a 1 dimensional array where each index represents the
        number of events in each bin
    """
    ct = np.zeros((bin_a, 1), dtype=np.int32)
    for x in np.arange(arr_a.size):
        ct[arr_a[x]] += 1
    return ct


@nb.jit(
    nopython=True, 
    nogil=True, 
    cache=True, 
    fastmath=True)
def hist2D(
        arr_a: np.ndarray, 
        arr_b: np.ndarray, 
        bin_a: int, 
        bin_b: int) -> np.ndarray:
    """calculates the cluster intersections between two arrays

    e.g. ct[0][2]: 
    the intersection of `arr_a` bin 0 & `arr_b` bin 2

    Parameters
	----------
    arr_a: np.ndarray
        a 1D array where each value represents a the cluster
        identity of a specific gene
    arr_b: np.ndarray
        a 1D array where each value represents a the cluster
        identity of a specific gene
    bin_a: np.ndarray
        the number of clusters in array 1. Should be equivalent
        to the maximum value in `arr_a` + 1
    bin_b: np.ndarray
        the number of clusters in array 1. Should be equivalent
        to the maximum value in `arr_b` + 1

    Returns
	-------
    np.ndarray
        a 2 dimensional array where each index represents
        the intersection of the cluster identities in
        `arr_a` and `arr_b`
    """
    assert arr_a.size == arr_b.size
    ct = np.zeros((bin_a, bin_b), dtype=np.int32)
    for x in np.arange(arr_a.size):
        ct[arr_a[x]][arr_b[x]] += 1
    return ct

@nb.jit(
    nopython=True, 
    nogil=True, 
    cache=True, 
    fastmath=True)
def hist3D(
        arr_a: np.ndarray, 
        arr_b: np.ndarray, 
        arr_c: np.ndarray, 
        bin_a: int, 
        bin_b: int, 
        bin_c: int) -> np.ndarray:
    """calculates the bin intersections between three arrays

    e.g. ct[0][2][1]: 
    the intersection of `arr_a` bin 0 & `arr_b` bin 2 & `arr_c` bin 1

    Parameters
	----------
    arr_a: np.ndarray
        a 1D array where each value represents a the cluster
        identity of a specific gene
    arr_b: np.ndarray
        a 1d array where each value represents a the cluster
        identity of a specific gene
    arr_c: np.ndarray
        a 1d array where each value represents a the cluster
        identity of a specific gene
    bin_a: np.ndarray
        the number of clusters in array 1. Should be equivalent
        to the maximum value in `arr_a` + 1
    bin_b: np.ndarray
        the number of clusters in array 1. Should be equivalent
        to the maximum value in `arr_b` + 1
    bin_c: np.ndarray
        the number of clusters in array 1. Should be equivalent
        to the maximum value in `arr_c` + 1

    Returns
	-------
    np.ndarray
        a 3 dimensional array where each index represents
        the intersection of the cluster identities in
        `arr_a` and `arr_b` and `arr_c`
    """
    assert arr_a.size == arr_b.size
    assert arr_b.size == arr_c.size
    ct = np.zeros((bin_a, bin_b, bin_c), dtype=np.int32)
    for x in np.arange(arr_a.size):
        ct[arr_a[x]][arr_b[x]][arr_c[x]] += 1
    return ct



"""Test information calculations
"""

import pytest
import numpy as np
from pypage.information import (
        entropy,
        conditional_entropy,
        joint_entropy,
        joint_entropy_3d,
        mutual_information,
        conditional_mutual_information)

EPSILON = 1e-9


def paired_arrays(
        size=1e4, 
        overlap=0.9, 
        bin_sizes=[0.05, 0.85, 0.05]):
    """janky solution to generating paired pseudo-bins and pseudo-ontologies.
    """

    size = int(size)
    bin_sizes = np.array(bin_sizes)
    bin_frac = bin_sizes / bin_sizes.sum()
    bin_sizes = (size * bin_sizes).astype(int)
    
    # create bins
    bins = np.random.choice(bin_sizes.size, p=bin_frac, size=size)
    
    # create ontology with overlap to first and last bin
    sig_max = np.max(bins)
    sig_min = np.min(bins)
    sig_choices = np.isin(bins, [sig_max, sig_min])
    sig_idx = np.flatnonzero(sig_choices)
    sig_total = sig_choices.sum()
    ontology = np.zeros(size, dtype=np.int32)
    mask = np.random.choice(sig_idx, size=int(overlap * sig_total), replace=False)
    ontology[mask] = 1

    x_bins = bins.max() + 1
    y_bins = ontology.max() + 1
    
    return (
        bins.astype(np.int32), 
        ontology.astype(np.int32), 
        x_bins, 
        y_bins)


def test_conditional_entropy():
    """test conditions of conditional entropy
    
    H(Y|X) 
        = H(X,Y) - H(X)
        = H(X|Y) - H(X) + H(Y)
    """
    x, y, x_bins, y_bins = paired_arrays()

    h_x = entropy(x, x_bins)
    h_y = entropy(y, y_bins)
    h_xy = joint_entropy(x, y, x_bins, y_bins)
    h_yGx = conditional_entropy(y, x, y_bins, x_bins)
    h_xGy = conditional_entropy(x, y, x_bins, y_bins)

    exp1 = h_xy - h_x
    exp2 = h_xGy - h_x + h_y

    assert np.abs(h_yGx - exp1) < EPSILON
    assert np.abs(h_yGx - exp2) < EPSILON

def test_mutual_information():
    """
    test alternative expressions of mutual information
    I(X;Y) 
        = H(X) - H(X|Y)
        = H(Y) - H(Y|X)
        = H(X) + H(Y) - H(X,Y)
        = H(X,Y) - H(X|Y) - H(Y|X)
        = I(Y;X)
    """
    x, y, x_bins, y_bins = paired_arrays()
    i_xy = mutual_information(x, y, x_bins, y_bins)
    i_yx = mutual_information(y, x, y_bins, x_bins)
    
    h_x = entropy(x, x_bins)
    h_y = entropy(y, y_bins)

    h_xGy = conditional_entropy(x, y, x_bins, y_bins)
    h_yGx = conditional_entropy(y, x, y_bins, x_bins)
    h_xy = joint_entropy(x, y, x_bins, y_bins)

    exp1 = h_x - h_xGy
    exp2 = h_y - h_yGx
    exp3 = h_x + h_y - h_xy
    exp4 = h_xy - h_xGy - h_yGx

    assert np.abs(i_xy - exp1) < EPSILON
    assert np.abs(i_xy - exp2) < EPSILON
    assert np.abs(i_xy - exp3) < EPSILON
    assert np.abs(i_xy - exp4) < EPSILON
    assert np.abs(i_xy - i_yx) < EPSILON

def test_joint_entropy():
    """test conditions of joint entropy using conditional and mutual information
    H(X,Y)
        = H(X) + H(Y) - I(X;Y)
        = H(X|Y) + H(Y|X) + I(X;Y)
    """
    x, y, x_bins, y_bins = paired_arrays()
    i_xy = mutual_information(x, y, x_bins, y_bins)
    h_xy = joint_entropy(x, y, x_bins, y_bins)
    
    h_x = entropy(x, x_bins)
    h_y = entropy(y, y_bins)
    h_xGy = conditional_entropy(x, y, x_bins, y_bins)
    h_yGx = conditional_entropy(y, x, y_bins, x_bins)
    
    exp1 = h_x + h_y - i_xy
    exp2 = h_xGy + h_yGx + i_xy

    assert np.abs(h_xy - exp1) < EPSILON
    assert np.abs(h_xy - exp2) < EPSILON

def test_conditional_mutual_information():
    """
    test alternative expressions of conditional mutual information
    I(X;Y|Z)
        = H(X,Z) + H(Y,Z) - H(X,Y,Z) - H(Z)
    """
    x, y, x_bins, y_bins = paired_arrays()
    z = y.copy()
    np.random.shuffle(z)
    z_bins = y_bins

    cmi = conditional_mutual_information(x, y, z, x_bins, y_bins, z_bins)

    h_xz = joint_entropy(x, z, x_bins, z_bins)
    h_yz = joint_entropy(y, z, y_bins, z_bins)
    h_xyz = joint_entropy_3d(x, y, z, x_bins, y_bins, z_bins)
    h_z = entropy(z, z_bins)

    exp1 = h_xz + h_yz - h_xyz - h_z

    assert np.abs(cmi - exp1) < EPSILON



"""Testing an example usage of the PAGE algorithm
"""

import pytest 
import pandas as pd
from pypage import (
    PAGE,
    ExpressionProfile,
    GeneSets)

@pytest.fixture()
def load_expression():
    frame = pd.read_csv(
        "example_data/AP2S1.tab.gz", 
        sep="\t",
        header=None,
        names=["gene", "bin"])

    return ExpressionProfile(
            frame.iloc[:, 0], 
            frame.iloc[:, 1], 
            is_bin=True)

@pytest.fixture()
def load_ontology():
    frame = pd.read_csv(
            "example_data/GO_BP_2021_index.txt.gz",
            sep="\t",
            header=None,
            names=["gene", "pathway"])

    return GeneSets(
            frame.iloc[:, 0],
            frame.iloc[:, 1])


def test_run(load_expression, load_ontology):
    p = PAGE(
        load_expression, 
        load_ontology, 
        n_shuffle=5, 
        k=2)
    results = p.run()




"""Data input and output handling for expression
"""


import numpy as np
from typing import Optional
from .accession_types import change_accessions


class ExpressionProfile:
    """Container for a Sample Expression Profile

    Attributes
    ==========
    genes: np.ndarray
        the sorted list of genes provided
    gene_indices: dict
        a dictionary where keys are gene names and values are indices
    bins: np.ndarray
        the sorted list of bins provided / calculated
    bin_indices: dict
        a dictionary where keys are bin names and values are indices
    n_genes: int
        the number of genes found
    n_bins: int
        the number of bins found / calculated
    bool_array: np.ndarray
        a (n_bins, n_genes) bit array representing the bin-position of each gene
    bin_array: np.ndarray
        a (n_genes, ) array where each value represents the bin index of that gene
    bin_sizes: np.ndarray
        a (n_bins, ) array where each value represents the number of genes in that bin index
    bin_ranges: Optional[np.ndarray]
        a (n_bins, ) array describing the ranges of the scores for each bin. only present if `is_bin` is True

    Methods
    -------
    get_gene_subset:
        returns a subset of the `bin_array` for the provided gene list
    """
    def __init__(
            self,
            genes: np.ndarray,
            expression: np.ndarray,
            is_bin: bool = False,
            n_bins: Optional[int] = 10):
        """
        Parameters
        ==========
        genes: np.ndarray
            The array with the gene names
        expression: np.ndarray
            The array representing either the continuous expression value
            of a specific gene, or the bin/cluster that gene belongs to.
        is_bin: bool
            whether the provided dataframe is prebinned.
        n_bins: int
        """
        self.modified = False
        self._is_bin = is_bin
        self._validate_inputs(genes, expression)
        self._process_input(genes, expression)
        self.n_bins = n_bins

    def _process_input(self,
                       x: np.ndarray,
                       y: np.ndarray):
        """
        sets the attributes accordingly to the inputs
        Parameters
        ----------
        x: np.ndarray
            input genes
        y: np.ndarray
            input expression
        """
        self.genes = np.array(x)
        self.n_genes = self.genes.size
        self.raw_expression = np.array(y)
        notnan_mask = ~ np.isnan(self.raw_expression)
        if len(self.raw_expression.shape) == 2:
            notnan_mask = notnan_mask.any(0)
            self.genes = self.genes[notnan_mask]
            self.raw_expression = self.raw_expression[:, notnan_mask]
        else:
            self.genes = self.genes[notnan_mask]
            self.raw_expression = self.raw_expression[notnan_mask]

    def _validate_inputs(
            self,
            x: np.ndarray,
            y: np.ndarray):
        """validates inputs are as expected
        """
        assert x.size > 0, \
            "provided array must not be empty"
        assert x.shape[0] == y.shape[-1], \
            "genes and expression/bin arrays must be equally shaped"

    def _discretize(self,
                    inp_array: np.ndarray,
                    bins: int,
                    noise_std: Optional[float] = 0.000000001):
        """
        discretizes the expression profile
        Parameters
        ----------
        inp_array: np.ndarray
        bins: int
        noise_std: float
        Returns
        -------
        np.ndarray
            discretized expression profile
        """

        length = len(inp_array)
        to_discr = inp_array + np.random.normal(0, noise_std, length)

        bins_for_discr = np.interp(np.linspace(0, length, bins + 1),
                                   np.arange(length),
                                   np.sort(to_discr))
        bins_for_discr[-1] += 1  # otherwise numpy creates one extra bin with only 1 point
        digitized = np.digitize(to_discr, bins_for_discr)
        digitized = digitized - 1
        return digitized

    def get_gene_subset(
            self,
            gene_subset: np.ndarray) -> np.ndarray:
        """
        Index the bin-array for the required gene subset.
        Expects the subset to be sorted

        Parameters
        ==========
        gene_subset: np.ndarray
            a list of genes to subset the bin_array to

        Returns
        =======
        np.ndarray
            the bin_array subsetted to the indices of the `gene_subset`
        """

        idxs = [np.where(self.genes == gene)[0][0] for gene in gene_subset]

        if len(self.raw_expression.shape) == 1:
            sub_expression = self.raw_expression[idxs]
            bin_array = self._discretize(sub_expression, self.n_bins)
        else:
            sub_expression = self.raw_expression[:, idxs]
            bin_array = np.apply_along_axis(lambda x: self._discretize(x, self.n_bins), 0, sub_expression)

        self.bin_array = bin_array
        self.modified = True
        return self.bin_array

    def convert_from_to(self,
                        input_format: str,
                        output_format: str,
                        species: Optional[str] = 'human'):
        """
        A function which changes accessions
        Parameters
        ----------
        input_format
            input accession type, takes 'enst', 'ensg', 'refseq', 'entrez', 'gs', 'ext'
        output_format
            output accession type, takes 'enst', 'ensg', 'refseq', 'entrez', 'gs', 'ext'
        species
            analyzed species, takes either 'human' or 'mouse'
        """
        if input_format in ['ensg', 'enst', 'refseq']:  # for ex., remove '.1' from 'ENSG00000128016.1'
            self.genes = np.array([gene.split('.')[0] for gene in self.genes])
        self.genes = change_accessions(self.genes,
                                       input_format,
                                       output_format,
                                       species)

    def reset(self):
        self.modified = False

    def __repr__(
            self) -> str:
        s = ""
        s += "Expression Profile\n"
        s += f">> num_genes: {self.n_genes}\n"
        s += f">> num_bins: {self.n_bins}\n"
        return s


from setuptools import setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
   name='bio-pypage',
   version='0.1.7',
   description='Python implementation of the PAGE algorithm',
   author='Artemy Bakulin, Noam Teyssier',
   long_description=long_description,
   long_description_content_type="text/markdown",
   url='https://github.com/goodarzilab/pypage',
   author_email="logic2000.bakulin@gmail.com",
   packages=['pypage', 'pypage.io'],
   install_requires=[
       'numpy',
       'pandas',
       'numba',
       'tqdm',
       'scipy',
       "pytest",
       "pybiomart",
       "matplotlib"],
   classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],)



"""Utility functions
"""

import numpy as np
import numba as nb
from scipy.stats import hypergeom
from .hist import (
        hist1D,
        hist2D,
        hist3D)

@nb.jit(nopython=True)
def shuffle_bin_array(
        bin_array: np.ndarray) -> np.ndarray:
    """Shuffles the bin identities for a provided bin array

    Parameters
    ----------
    bin_array: np.ndarray
        any type of array to be shuffled

    Returns
    -------
    np.ndarray
        a shuffled copy of `bin_array`

    Examples
    --------
    >>> a = np.random.choice(3, size=5)
    >>> b = shuffle_bin_array(a)
    >>> print(a)
    [0 1 4 1 2]
    >>> print(b)
    [2 4 1 1 0]
    """
    shuf = bin_array.copy()
    np.random.shuffle(shuf)
    return shuf


@nb.jit(nopython=True)
def empirical_pvalue(
        array: np.ndarray,
        value: float) -> float:
    """Calculates the empirical pvalue of a value given a distribution

    Parameters
    ----------
    array: np.ndarray
        an array of values representing the distribution to test against
    value: float
        a value to calculate the empirical p-value against `array`

    Returns
    -------
    float
        the empirical pvalue of the value against the provided array

    Examples
    --------
    >>> a = np.random.random(100000)
    >>> p = 0.95
    >>> empirical_pvalue(a, p)
    0.04974
    """
    m = array.max()
    if value > m:
        return 0.
    else:
        return np.mean(array >= value)


def hypergeometric_test(
        e_bins: np.ndarray, 
        o_bool: np.ndarray) -> np.ndarray:
    """
    
    Variable names from supplemental [1]_ translated to scipy documentation [2]_

        ``n`` : number of genes in the pathway
        
        ``k`` : number of genes in the bin which are also in the pathway
        
        ``N`` : number of genes in the bin
        
        ``M`` : total number of genes

    Hypergeometric Form:

    .. math::

        p(k, M, n, N) = \\frac {\\binom{n}{k} \\binom{M-n}{N-k}} {\\binom{M}{N}}

    Parameters
    ----------
    e_bins: np.ndarray
        the 1D array where each element represents the bin index of a particular gene
    o_bool: np.ndarray
        the 1D array where each element represents whether the gene is found in the present pathway

    Results
    -------
    np.ndarray
        a 2D array of shape (2, n_bins). 
        the first array represents the p-value of overrepresentation
        the second array represents the p-value of underrepresentation

    Notes
    -----
    .. [1] PAGE Supplemental Methods. https://ars.els-cdn.com/content/image/1-s2.0-S1097276509008570-mmc1.pdf
    .. [2] scipy hypergeometric documentation. https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.hypergeom.html
    """
    n_bins = e_bins.max() + 1
    ct = hist2D(e_bins, o_bool, n_bins, 2)
    n = o_bool.sum()
    k = ct[:, 1]
    N = ct.sum(axis=1)
    M = o_bool.size

    # parameterize hypergeometric distribution
    hg = hypergeom(M, n, N)
    
    # measure overrepresentation
    sf = hg.sf(k-1)
    
    # measure underrepresentation
    cdf = hg.cdf(k)
    
    return np.stack([sf, cdf])


def benjamini_hochberg(
        p: np.ndarray) -> np.ndarray:
    """
    Benjamini-Hochberg p-value correction for multiple hypothesis testing. 

    Parameters
    ----------
    p: np.ndarray
        an array of p-values to correct

    Returns
    -------
    np.ndarray
        an array of adjusted p-values

    Examples
    --------
    >>> pvals = np.random.random(1000000)
    >>> pvals[:10] = np.random.random(10) * 1e-40
    >>> qvals = benjamini_hochberg(pvals)
    >>> qvals[qvals < 0.05].size
    10
    """
    p = np.asfarray(p)
    by_descend = p.argsort()[::-1]
    by_orig = by_descend.argsort()
    steps = float(len(p)) / np.arange(len(p), 0, -1)
    q = np.minimum(1, np.minimum.accumulate(steps * p[by_descend]))
    return q[by_orig]



"""Testing an example usage of the PAGE algorithm with the simulated data
"""

import pytest
import pandas as pd
from pypage import (
    PAGE,
    ExpressionProfile,
    GeneSets)

@pytest.fixture()
def load_expression():
    df = pd.read_csv('example_data/simulated_expr.csv.gz', header=0, index_col=0, compression='gzip')
    exp = ExpressionProfile(df.iloc[:, 0],
                            df.iloc[:, 1],
                            n_bins=10)
    return exp

@pytest.fixture()
def load_ontology():
    frame = pd.read_csv('example_data/simulated_df.csv.gz', header=0, index_col=0, compression='gzip')
    ont = GeneSets(
        frame.iloc[:, 0],
        frame.iloc[:, 1],
        n_bins=6)
    return ont


def test_run(load_expression, load_ontology):
    p = PAGE(
        load_expression,
        load_ontology,
        n_shuffle=100,
        k=10,
        alpha=0.01)
    results, hm = p.run()

    positive = set(['upregulated_' + str(i) for i in range(50)] + ['downregulated_' + str(i) for i in range(50)])
    negative = set(['random_' + str(i) for i in range(1000)])
    count_true_positive = lambda series: len(set(series) & positive)
    count_true_negative = lambda series: len(negative - set(series))
    tpr = count_true_positive(results['pathway']) / 100
    tnr = count_true_negative(results['pathway']) / 1000
    print('tpr: %.2f\ntnr: %.2f' % (tpr, tnr))
    assert tpr > .9
    assert tnr > .9



"""Testing Ontology IO
"""


import numpy as np
from pypage import GeneSets


N_GENES=1000
N_PATHWAYS=50
T=100


def get_ontology() -> (np.ndarray, np.ndarray):
    genes = np.array([f"g.{g}" for g in np.arange(N_GENES)])
    pathways = np.random.choice(N_PATHWAYS, size=N_GENES)
    return genes, pathways


def filter_assertions(
        ont: GeneSets,
        mask: np.ndarray, 
        unique_pathways: np.ndarray, 
        counts: np.ndarray):

    filtered_pathways = unique_pathways[mask]
    filtered_counts = counts[mask]

    assert ont.pathways.size == filtered_pathways.size
    assert np.all(ont.pathways == filtered_pathways)
    assert np.all(ont.pathway_sizes == filtered_counts)


def test_load_ontology():
    for _ in np.arange(T):
        genes, pathways = get_ontology()
        ont = GeneSets(genes, pathways)


def test_load_ontology_assertion():
    for i in np.arange(T):
        genes, pathways = get_ontology()

        if i % 2 == 0:
            genes = genes[np.random.random(N_GENES) < 0.3]
        else:
            pathways = pathways[np.random.random(N_GENES) < 0.3]

        try:
            ont = GeneSets(genes, pathways)
        except AssertionError:
            continue

        assert False


def test_ontology_min_filtering():
    for _ in np.arange(T):
        min_size = np.random.choice(np.arange(0, 3))
        genes, pathways = get_ontology()
        ont = GeneSets(genes, pathways)

        unique_pathways, counts = np.unique(pathways, return_counts=True)
        mask = counts >= min_size
        ont.filter_pathways(min_size=min_size)

        filter_assertions(ont, mask, unique_pathways, counts)


def test_ontology_max_filtering():
    for _ in np.arange(T):
        max_size = np.random.choice(np.arange(8, 10))
        genes, pathways = get_ontology()
        ont = GeneSets(genes, pathways)

        unique_pathways, counts = np.unique(pathways, return_counts=True)
        mask = counts <= max_size
        ont.filter_pathways(max_size=max_size)
        filter_assertions(ont, mask, unique_pathways, counts)


def test_ontology_bandpass_filtering():
    for _ in np.arange(T):
        min_size = np.random.choice(np.arange(0, 3))
        max_size = np.random.choice(np.arange(8, 10))
        genes, pathways = get_ontology()
        ont = GeneSets(genes, pathways)

        unique_pathways, counts = np.unique(pathways, return_counts=True)
        mask = (counts >= min_size) & (counts <= max_size)
        ont.filter_pathways(min_size=min_size, max_size=max_size)
        filter_assertions(ont, mask, unique_pathways, counts)


def test_ontology_null_filtering():
    for _ in np.arange(T):
        genes, pathways = get_ontology()
        ont = GeneSets(genes, pathways)
        
        unique_pathways, counts = np.unique(pathways, return_counts=True)
        mask = np.ones(unique_pathways.size, dtype=bool)
        ont.filter_pathways()
        filter_assertions(ont, mask, unique_pathways, counts)


def test_ontology_min_assertion():
    for _ in np.arange(T):
        genes, pathways = get_ontology()
        ont = GeneSets(genes, pathways)

        try:
            ont.filter_pathways(min_size = -1)
        except AssertionError:
            continue

        assert False


def test_ontology_max_assertion():
    for _ in np.arange(T):
        genes, pathways = get_ontology()
        ont = GeneSets(genes, pathways)

        try:
            ont.filter_pathways(min_size = 10, max_size=5)
        except AssertionError:
            continue

        assert False


def test_read_annotation_file():
    ont = GeneSets(ann_file='example_data/hg38_cistrome_index.txt.gz')
    return ont


def test_convert_from_to():
    ont = GeneSets(ann_file='example_data/hg38_cistrome_index.txt.gz')
    ont.convert_from_to('ensg', 'gs', 'human')
    return ont